from datetime import datetime
from io import BytesIO
from pathlib import Path
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    TYPE_CHECKING,
    Tuple,
    Union,
    cast,
)
from urllib.parse import quote_plus
from pararamio.constants import POSTS_LIMIT
from pararamio.exceptions import (
    PararamioLimitExceededException,
    PararamioMethodNotAllowed,
    PararamioRequestException,
    PararamioValidationException,
)
from pararamio.utils.helpers import (
    encode_chats_ids,
    format_datetime,
    join_ids,
    parse_iso_datetime,
)
from pararamio.post import Post
from ._base import BaseLoadedAttrPararamObject

if TYPE_CHECKING:
    from pararamio.client import Pararamio
    from pararamio.file import File
    from pararamio._types import FormatterT, QuoteRangeT

__all__ = ('Chat',)

ATTR_FORMATTERS: 'FormatterT' = {
    'time_edited': parse_iso_datetime,
    'time_created': parse_iso_datetime,
    'user_time_edited': parse_iso_datetime,
}


def check_result(result: dict) -> bool:
    return 'chat_id' in result


class Chat(BaseLoadedAttrPararamObject):
    id: int
    title: str
    history_mode: str
    description: Optional[str]
    posts_count: int
    pm: bool
    e2e: bool
    time_created: datetime
    time_updated: datetime
    time_edited: Optional[datetime]
    author_id: int
    two_step_required: bool
    org_visible: bool
    organization_id: Optional[int]
    posts_live_time: Optional[int]
    allow_api: bool
    read_only: bool
    tnew: bool
    adm_flag: bool
    custom_title: Optional[str]
    is_favorite: bool
    inviter_id: Optional[int]
    tshow: bool
    user_time_edited: datetime
    history_start: int
    pinned: List[int]
    thread_groups: List[int]
    thread_users: List[int]
    thread_admins: List[int]
    thread_users_all: List[int]
    last_msg_author_id: Optional[int]
    last_msg_author: str
    last_msg_bot_name: str
    last_msg_text: str
    last_msg: str
    last_read_post_no: int
    thread_guests: List[int]
    _data: Dict[str, Any]
    _client: 'Pararamio'

    def __init__(
        self,
        client: 'Pararamio',
        id: Optional[int] = None,
        load_on_key_error: bool = True,
        **kwargs: Any,
    ) -> None:
        if id is None:
            id = kwargs.get('chat_id', None)
            if id is None:
                id = kwargs['thread_id']
        self.id = int(id)
        self._data = {}
        if kwargs:
            self._data = {**kwargs, 'id': id}
        self._load_on_key_error = load_on_key_error
        self._client = client

    def __str__(self) -> str:
        title = self._data.get('title', '')
        id_ = self.id or ''
        return f'{id_} - {title}'

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Chat):
            return id(other) == id(self)
        return self.id == other.id

    def __contains__(self, item: 'Post') -> bool:
        if isinstance(item, Post):
            return item.chat == self
        return False

    @property
    def client(self) -> 'Pararamio':
        return self._client

    def load(self) -> 'Chat':
        if self.id is None:
            raise PararamioMethodNotAllowed(f'Load is not allow for new {self.__class__.__name__}')
        chats = self.load_chats(self._client, [self.id])
        if len(chats) != 1:
            raise PararamioRequestException(f'failed to load data for chat id {self.id}')
        self._data = chats[0]._data
        return self

    def edit(self, **kwargs) -> None:
        """
        Updates the attributes of a chat instance with the provided keyword arguments.

        Parameters:
          kwargs: Arbitrary keyword arguments specifying the attributes to update.

        Returns:
          None

        Raises:
          Various exceptions based on the response from the API PUT request.
        """
        url = f'/core/chat/{self.id}'
        check_result(self._client.api_put(url, data=kwargs))

    def transfer(self, org_id: int) -> bool:
        url = f'/core/chat/{self.id}/transfer/{org_id}'
        return check_result(self._client.api_post(url, {}))

    def delete(self):
        url = f'/core/chat/{self.id}'
        return check_result(self._client.api_delete(url))

    def hide(self) -> bool:
        url = f'/core/chat/{self.id}/hide'
        return check_result(self._client.api_post(url))

    def show(self) -> bool:
        url = f'/core/chat/{self.id}/show'
        return check_result(self._client.api_post(url))

    def favorite(self) -> bool:
        url = f'/core/chat/{self.id}/favorite'
        return check_result(self._client.api_post(url))

    def unfavorite(self) -> bool:
        url = f'/core/chat/{self.id}/unfavorite'
        return check_result(self._client.api_post(url))

    def enter(self) -> bool:
        url = f'/core/chat/{self.id}/enter'
        return check_result(self._client.api_post(url))

    def quit(self) -> bool:
        url = f'/core/chat/{self.id}/quit'
        return check_result(self._client.api_post(url))

    def set_custom_title(self, title: str) -> bool:
        url = f'/core/chat/{self.id}/custom_title'
        return check_result(self._client.api_post(url, {'title': title}))

    def add_users(self, ids: List[int]) -> bool:
        url = f'/core/chat/{self.id}/user/{join_ids(ids)}'
        return check_result(self._client.api_post(url))

    def delete_users(self, ids: List[int]) -> bool:
        url = f'/core/chat/{self.id}/user/{join_ids(ids)}'
        return check_result(self._client.api_delete(url))

    def add_admins(self, ids: List[int]) -> bool:
        url = f'/core/chat/{self.id}/admin/{join_ids(ids)}'
        return check_result(self._client.api_post(url))

    def delete_admins(self, ids: List[int]) -> bool:
        url = f'/core/chat/{self.id}/admin/{join_ids(ids)}'
        return check_result(self._client.api_delete(url))

    def add_groups(self, ids: List[int]) -> bool:
        url = f'/core/chat/{self.id}/group/{join_ids(ids)}'
        return check_result(self._client.api_post(url))

    def delete_groups(self, ids: List[int]) -> bool:
        url = f'/core/chat/{self.id}/group/{join_ids(ids)}'
        return check_result(self._client.api_delete(url))

    def _load_posts(self, start_post_no: int = -50, end_post_no: int = -1) -> List['Post']:
        url = f'/msg/post?chat_id={self.id}&range={start_post_no}x{end_post_no}'
        if (start_post_no < 0 <= end_post_no) or (start_post_no >= 0 > end_post_no):
            raise PararamioValidationException(
                'start_post_no and end_post_no can only be negative or positive at the same time'
            )
        if 0 > start_post_no > end_post_no:
            raise PararamioValidationException(
                'range start_post_no must be greater then end_post_no'
            )
        if 0 <= start_post_no > end_post_no:
            raise PararamioValidationException(
                'range start_post_no must be smaller then end_post_no'
            )
        _absolute = abs(end_post_no - start_post_no)
        if start_post_no < 0:
            _absolute = +1
        if _absolute >= POSTS_LIMIT:
            raise PararamioLimitExceededException(f'max post load limit is {POSTS_LIMIT - 1}')
        res = self._client.api_get(url).get('posts', [])
        if not res:
            return []
        return [Post(chat=self, **post) for post in res]

    def _lazy_posts_loader(
        self, start_post_no: int = -50, end_post_no: int = -1
    ) -> Iterable['Post']:
        if (start_post_no < 0 <= end_post_no) or (start_post_no >= 0 > end_post_no):
            raise PararamioValidationException(
                'start_post_no and end_post_no can only be negative or positive at the same time'
            )
        if 0 > start_post_no > end_post_no:
            raise PararamioValidationException(
                'range start_post_no must be greater then end_post_no'
            )
        if 0 <= start_post_no > end_post_no:
            raise PararamioValidationException(
                'range start_post_no must be smaller then end_post_no'
            )
        absolute = abs(end_post_no - start_post_no)
        start, end = start_post_no, end_post_no
        if absolute > POSTS_LIMIT:
            end = start_post_no + POSTS_LIMIT - 1
        posts = iter(self._load_posts(start, end))
        counter = 0
        for _ in range(start_post_no, end_post_no):
            try:
                yield next(posts)
            except StopIteration:
                counter += 1
                res = self._load_posts(start + POSTS_LIMIT * counter, end + POSTS_LIMIT * counter)
                if not res:
                    return
                posts = iter(res)

    def posts(self, start_post_no: int = -50, end_post_no: int = -1) -> List['Post']:
        return self._load_posts(start_post_no=start_post_no, end_post_no=end_post_no)

    def read_status(self, post_no: int) -> bool:
        url = f'/msg/lastread/{self.id}'
        res = self._client.api_post(url, {'post_no': post_no})
        if not res:
            return False
        if 'post_no' in res:
            self._data['last_read_post_no'] = res['post_no']
        if 'posts_count' in res:
            self._data['posts_count'] = res['posts_count']
        return True

    def post(
        self,
        text: str,
        quote_range: Optional['QuoteRangeT'] = None,
        reply_no: Optional[int] = None,
    ) -> 'Post':
        if self.id is None:
            raise ValueError('can not post file to new chat')
        return Post.create(
            self,
            text=text,
            reply_no=reply_no,
            quote=cast(str, quote_range['text']) if quote_range else None,
        )

    def upload_file(
        self,
        file: Union[str, BytesIO, Path],
        filename: Optional[str] = None,
        reply_no: Optional[int] = None,
        quote_range: Optional[str] = None,
    ) -> 'File':
        if self.id is None:
            raise ValueError('can not upload file to new chat')
        if not isinstance(file, (str, Path)) and not filename:
            raise PararamioValidationException('can not determine filename for BinaryIO')
        return self._client.upload_file(
            file=file,
            chat_id=self.id,
            filename=filename,
            reply_no=reply_no,
            quote_range=quote_range,
        )

    @classmethod
    def load_chats(cls, client: 'Pararamio', ids: List[int]) -> List['Chat']:
        url = f'/core/chat?ids={join_ids(ids)}'
        res = client.api_get(url)
        if res and 'chats' in res:
            return [cls(client, **data) for data in client.api_get(url).get('chats', [])]
        raise PararamioRequestException(
            f'failed to load data for chats ids: {",".join(map(str, ids))}'
        )

    @classmethod
    def post_search(
        cls,
        client: 'Pararamio',
        q: str,
        order_type: str = 'time',
        page: int = 1,
        chat_id: Optional[int] = None,
        limit: Optional[int] = POSTS_LIMIT,
    ) -> Tuple[int, Iterable['Post']]:
        if not limit:
            limit = POSTS_LIMIT
        url = f'/posts/search?q={quote_plus(q)}&order_type={order_type}&page={page}&limit={limit}'
        if chat_id is not None:
            url += f'&th_id={chat_id}'

        res = client.api_get(url)
        if 'posts' not in res:
            raise PararamioRequestException('failed to perform search')
        created_chats = {}

        def create_post(data):
            nonlocal created_chats
            _chat_id = data['thread_id']
            post_no = data['post_no']
            if _chat_id not in created_chats:
                created_chats[_chat_id] = cls(client, id=_chat_id)
            return Post(created_chats[_chat_id], post_no=post_no)

        return res['count'], map(create_post, res['posts'])

    @classmethod
    def create(
        cls,
        client: 'Pararamio',
        title: str,
        description: str = '',
        users: Optional[List[int]] = None,
        groups: Optional[List[int]] = None,
        **kwargs,
    ) -> 'Chat':
        """

        Creates a new chat instance in the Pararamio application.

        Args:
            cls: The class itself (implicit first argument for class methods).
            client (Pararamio): An instance of the Pararamio client.
            title (str): The title of the chat.
            description (str, optional): A description of the chat. Default is an empty string.
            users (Optional[List[int]], optional): A list of user IDs to be added to the chat.
                                                   Default is None.
            groups (Optional[List[int]], optional): A list of group IDs to be added to the chat.
                                                    Default is None.
            **kwargs: Additional keyword arguments to be included in the chat creation data.

        Returns:
            Chat: An instance of the Chat class representing the newly created chat.
        """
        if users is None:
            users = []
        if groups is None:
            groups = []
        data = {
            'title': title,
            'description': description,
            'users': users,
            'groups': groups,
            **kwargs,
        }

        res = client.api_post('/core/chat', data)
        id_: int = res['chat_id']
        return cls(client, id_)

    @classmethod
    def create_private_chat(cls, client: 'Pararamio', user_id: int) -> 'Chat':
        url = f'/core/chat/pm/{user_id}'
        res = client.api_post(url)
        id_: int = res['chat_id']
        return cls(client, id=id_)

    @staticmethod
    def sync_chats(
        client: 'Pararamio',
        chats_ids: List[Tuple[int, int, int]],
        sync_time: Optional[datetime] = None,
    ) -> Dict[str, Any]:
        url = '/core/chat/sync'
        data = {'ids': encode_chats_ids(chats_ids)}
        if sync_time:
            data['sync_time'] = format_datetime(sync_time)
        return client.api_post(url)



from datetime import datetime
from typing import Any, List, Optional, TYPE_CHECKING

from pararamio.exceptions import PararamNotFound
from pararamio.utils import format_datetime, format_or_none, parse_iso_datetime
from ._base import BaseLoadedAttrPararamObject

if TYPE_CHECKING:
    from pararamio.client import Pararamio
    from pararamio._types import FormatterT, QuoteRangeT

__all__ = ('DeferredPost',)
ATTR_FORMATTERS: 'FormatterT' = {
    'time_created': parse_iso_datetime,
    'time_sending': parse_iso_datetime,
}


class DeferredPost(BaseLoadedAttrPararamObject):
    id: int
    user_id: int
    chat_id: int
    text: str
    reply_no: Optional[int]
    time_created: datetime
    time_sending: datetime
    data: dict
    _data: dict
    _client: 'Pararamio'

    def __init__(self, client: 'Pararamio', id: int, load_on_key_error: bool = True, **kwargs):
        self._client = client
        self.id = id
        self._data = {'id': id, **kwargs}
        self._load_on_key_error = load_on_key_error

    def _get_formatted_attr(self, key: str) -> Any:
        try:
            try:
                return format_or_none(key, self._data, ATTR_FORMATTERS)
            except KeyError:
                return self._data.get('data', {})[key]
        except KeyError:
            if self._load_on_key_error:
                self.load()
                try:
                    return format_or_none(key, self._data, ATTR_FORMATTERS)
                except KeyError:
                    return self._data.get('data', {})[key]
            raise

    def __str__(self):
        text = self._data.get('text', None)
        if text is None:
            self.load()
            text = self._data['text']
        return text

    def load(self) -> 'DeferredPost':
        for post in self.get_deferred_posts(self._client):
            if post.id == self.id:
                self._data = post._data
                print(self._data)
                return self
        raise PararamNotFound(f'Deferred post with id {self.id} not found')

    def delete(self):
        url = f'/msg/deferred/{self.id}'
        self._client.api_delete(url)

    @classmethod
    def create(
        cls,
        client: 'Pararamio',
        chat_id: int,
        text: str,
        time_sending: datetime,
        reply_no: Optional[int] = None,
        quote_range: Optional['QuoteRangeT'] = None,
    ) -> 'DeferredPost':
        url = '/msg/deferred'
        data = {
            'chat_id': chat_id,
            'text': text,
            'time_sending': format_datetime(time_sending),
            'reply_no': reply_no,
            'quote_range': quote_range,
        }
        res = client.api_post(url, data)
        return cls(
            client,
            id=int(res['deferred_post_id']),
            chat_id=chat_id,
            data=data,
            time_sending=time_sending,
        )

    @classmethod
    def get_deferred_posts(cls, client: 'Pararamio') -> List['DeferredPost']:
        url = '/msg/deferred'
        res = client.api_get(url).get('posts', [])
        return [cls(client, **post) for post in res]



import os
import tempfile
import time
import unittest
from datetime import datetime, timedelta, timezone
from http.cookiejar import MozillaCookieJar
from unittest.mock import MagicMock, patch
import logging
from typing import List

from pararamio import Group
from pararamio.chat import Chat
from pararamio.client import Pararamio
from pararamio.deferred_post import DeferredPost
from pararamio.exceptions import (
    PararamioException,
    PararamioHTTPRequestException,
    PararamioValidationException,
)
from pararamio.post import Post
from pararamio.user import User
from pararamio.activity import ActivityAction
from pararamio.team import Team
from pararamio.bot import PararamioBot
from pararamio.poll import Poll
from tests.integrations._base import BasePararamioTest


RESOURCES_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), '../resources')
SKIP_AUTH_TESTS = bool(int(os.getenv('SKIP_AUTH_TESTS', '1')))


class PararamioClientTest(BasePararamioTest):
    def setUp(self):
        super().setUp()
        if int(os.environ.get('DEBUG', 0)):
            logging.basicConfig(level=logging.DEBUG)
        self.recipient_email = os.environ.get('PARARAMIO_RECIPIENT_EMAIL', '')
        self.recipient_chat_id = int(os.environ.get('PARARAMIO_RECIPIENT_CHAT_ID', 0))
        self.org_id = int(os.environ.get('PARARAMIO_ORG_ID', 0))
        self.bot = PararamioBot(os.environ.get('PARARAMIO_TEST_BOT_KEY', ''))
        if not self.recipient_email:
            self.skipTest('env variable PARARAMIO_RECIPIENT_EMAIL is not set')
        if not self.recipient_chat_id:
            self.skipTest('env variable PARARAMIO_RECIPIENT_CHAT_ID is not set')
        if not self.org_id:
            self.skipTest('env variable PARARAMIO_ORG_ID is not set')
        if not self.bot.key:
            self.skipTest('env variable PARARAMIO_TEST_BOT_KEY is not set')

    def test_001_login_without_cookie(self):
        if SKIP_AUTH_TESTS:
            self.skipTest('set env variable SKIP_AUTH_TESTS to 0 to run this test')
        client = Pararamio(**self.user1)
        self.assertIsNotNone(client.profile)

    def test_002_login_with_cookie(self):
        if SKIP_AUTH_TESTS:
            self.skipTest('set env variable SKIP_AUTH_TESTS to 0 to run this test')
        with tempfile.NamedTemporaryFile(suffix='.cookie') as tmp:
            MozillaCookieJar(tmp.name).save()
            client = Pararamio(**self.user1, cookie_path=tmp.name)
            self.assertIsNotNone(client.profile, 'Authentication failed')
            self.assertTrue(os.path.exists(tmp.name), 'Cookie path is not found')
            del client
            client = Pararamio(login='', password='', key='', cookie_path=tmp.name)
            self.assertIsNotNone(client.profile, 'Cookie authentication failed')

    def test_003_login_with_broken_cookie(self):
        if SKIP_AUTH_TESTS:
            self.skipTest('set env variable SKIP_AUTH_TESTS to 0 to run this test')
        with tempfile.NamedTemporaryFile(suffix='.cookie') as tmp:
            tmp.write('broken_cookie'.encode())
            client = Pararamio(**self.user1, cookie_path=tmp.name, ignore_broken_cookie=True)
            self.assertIsNotNone(client.profile, 'Authentication failed')
            self.assertTrue(os.path.exists(tmp.name), 'Cookie path is not found')

    def test_004_profile(self):
        self.assertEqual(self.client.profile.get('email'), self.user1['login'])

    def test_005_send_delete_file(self):
        d = {'size': 23050, 'type': 'image/png', 'filename': 'test.png'}
        post = None
        chat = Chat(self.client, self.recipient_chat_id)
        file = chat.upload_file(file=os.path.join(RESOURCES_PATH, 'test.png'))
        chat.load()
        post = Post(chat, post_no=chat.posts_count)
        self.assertEqual(post.file.guid, file.guid)
        self.assertEqual(post.file.filename, d['filename'])
        self.assertIsNotNone(file)
        file.delete()
        self.assertDictEqual(
            {'size': file.size, 'type': post.file._data['mime_type'], 'filename': file.filename}, d
        )

    def test_006_list_own_threads(self):
        chats = self.client.list_chats()
        self.assertIsNotNone(chats)
        chat = next(chats)
        chat_ = Chat(self.client, chat.id)
        chat_.load()
        self.assertTrue(chat == chat_)

    def test_007_list_posts(self):
        chat = Chat(self.client, self.recipient_chat_id)
        self.assertTrue(chat.posts(-10, -1))
        self.assertTrue(chat.posts(0, 1))
        chat.load()
        exception = False
        try:
            chat.posts(1, 0)
        except PararamioValidationException:
            exception = True
        self.assertTrue(exception, 'posts load with invalid start end position')
        exception = False
        try:
            chat.posts(-1, -10)
        except PararamioValidationException:
            exception = True
        self.assertTrue(exception, 'posts load with invalid start end position')

    def test_008_search_user(self):
        self.assertIsNotNone(self.client.search_user('tamtambottester2@gmail.com'))
        self.assertIsNotNone(self.client.search_user('tamtambottester2'))
        self.assertIsNotNone(self.client.search_user('tamtam_bot_tester2'))

    def test_009_reply_to_post(self):
        chat = Chat(self.client, self.recipient_chat_id)
        post = chat.post('test')
        reply = post.reply('ReplyTest')
        self.assertEqual(post.post_no, reply.reply_no)

    def test_010_search_group(self):
        self.assertIsNotNone(self.client.search_group('test'))

    def test_011_list_users(self):
        u = User(self.client, 58)
        u.load()

        self.assertTrue(u.unique_name)

    def test_012_chat_load(
        self,
    ):
        u = self.client.search_user(self.recipient_email)[0]
        chat = u.get_pm_thread()
        chat.load()
        pc = chat.posts_count
        u.post('text')
        chat.load()
        self.assertTrue(chat.posts_count == pc + 1)

    def test_013_read_status(self):
        u = self.client.search_user(self.recipient_email)[0]
        chat = u.get_pm_thread()
        chat.load()
        post = chat.post('test')
        chat.read_status(post.post_no + 1)
        self.assertTrue(chat.last_read_post_no == post.in_thread_no)

    def test_014_edit(self):
        u = self.client.search_user(self.recipient_email)[0]
        post = u.post('test')
        post.edit(text='TestTestStringString')
        chat = u.get_pm_thread()
        np = Post(chat, post.post_no)
        np.load()
        self.assertEqual('TestTestStringString', np.text)

    def test_015_send_and_delete_private_message(self):
        u = self.client.search_user(self.recipient_email)[0]
        post = u.post('test')
        self.assertIsNotNone(post)
        self.assertIsNotNone(post.delete())

    def test_016_send_and_delete_thread_message(self):
        chat = Chat(self.client, self.recipient_chat_id)
        post = chat.post('test')
        self.assertIsNotNone(post)
        self.assertIsNotNone(post.delete())

    def test_017_chat_methods(self):
        d = {
            'title': 'Test Create Chat',
            'users': [self.client.profile['id']],
            'pm': False,
            'member_ids': [self.client.profile['id']],
        }
        chat = Chat.create(self.client2, **d)
        self.assertTrue(
            Post(
                chat,
                post_no=1,
            )
            in chat
        )
        self.assertFalse(Post(Chat(self.client, id=100), post_no=1) in chat)
        self.assertTrue(
            chat.id in [ch.id for ch in self.client.list_chats()], 'Thread is not created'
        )
        chat.add_admins([self.client.profile['id']])
        chat.load()
        chat2 = Chat(self.client, chat.id).load()
        self.assertTrue(chat2.adm_flag, 'Admin is not set')
        time.sleep(5)
        chat.delete_admins([self.client.profile['id']])
        chat2.load()
        self.assertFalse(chat2.adm_flag, 'User is still admin')
        chat.delete()
        self.assertFalse(
            chat.id in [tr.id for tr in self.client.list_chats()], 'Thread is still exists'
        )

    def test_018_deferred_posts(self):
        time_sending = datetime.now(tz=timezone.utc) + timedelta(days=1)
        text = f'{time_sending.timestamp()}: test deferred posts !'
        result = DeferredPost.create(self.client, self.recipient_chat_id, text, time_sending)
        self.assertTrue(result)
        posts = DeferredPost.get_deferred_posts(self.client)
        post = [p for p in posts if p.text == text]
        self.assertTrue(bool(post), 'Post is empty')
        self.assertTrue(post[0].text == text, 'text is not match')
        self.assertTrue(int(post[0].time_sending.timestamp()) == int(time_sending.timestamp()))
        post[0].delete()

    def test_019_posts_tree(self):
        chat = Chat(self.client, self.recipient_chat_id, posts_count=100)
        with patch.object(
            Post,
            'rerere',
            return_value=[
                Post(chat, post_no=96, text='post no 96', reply_no=None),
                Post(chat, post_no=99, text='post no 99', reply_no=96),
                Post(chat, post_no=100, text='post no 100', reply_no=99, meta={}),
            ],
        ):
            with patch.object(
                chat,
                '_lazy_posts_loader',
                return_value=iter(
                    [
                        Post(chat, post_no=96, text='post no 96', reply_no=None),
                        Post(chat, post_no=97, text='post no 97', reply_no=None),
                        Post(chat, post_no=98, text='post no 98', reply_no=96),
                        Post(chat, post_no=99, text='post no 99', reply_no=96),
                    ]
                ),
            ):
                post = Post(chat, post_no=100, text='post no 100', reply_no=99, meta={})
                self.assertListEqual([96, 98, 99, 100], list(post.get_tree().keys()))
        many_fake_posts = [
            Post(chat, post_no=1, text='post no 1', reply_no=None),
            *[Post(chat, text=f'post no {i}', post_no=i, reply_no=i - 1) for i in range(2, 2000)],
            Post(chat, post_no=2000, text='post no 2000', reply_no=1999, meta={}),
        ]

        def _load_posts_side_effect(start, end):
            return iter(many_fake_posts[start:end])

        with patch.object(
            Post, 'rerere', return_value=[many_fake_posts[0], *many_fake_posts[1000:]]
        ):
            chat._load_posts = MagicMock(side_effect=_load_posts_side_effect)
            posts = list(many_fake_posts[-1].get_tree().keys())
            self.assertListEqual(
                [many_fake_posts[0].post_no, *[p.post_no for p in many_fake_posts[1000:]]], posts
            )

        with patch.object(Post, 'rerere', return_value=[]):
            chat._load_posts = MagicMock(side_effect=lambda a, b: [])
            self.assertListEqual([2000], list(many_fake_posts[-1].get_tree().keys()))

    def test_020_send_and_delete_message_by_email(self):
        text = 'test by mail'
        try:
            self.client.post_private_message_by_user_email(self.recipient_email + '1111', text)
            self.assertTrue(False, 'the message cannot be sent to a non-existent email')
        except PararamioException:
            pass
        post = self.client.post_private_message_by_user_email(self.recipient_email, text)
        self.assertIsNotNone(post)
        self.assertTrue(post.text == text)
        self.assertIsNotNone(post.delete())

    def test_021_send_and_delete_message_by_id(self):
        text = 'test by mail'
        try:
            self.client.post_private_message_by_user_email(-1, text)
            self.assertTrue(False, 'the message cannot be sent to a non-existent id')
        except PararamioException:
            pass
        post = self.client.post_private_message_by_user_id(self.client2.profile['id'], text)
        self.assertIsNotNone(post)
        self.assertTrue(post.text == text)
        self.assertIsNotNone(post.delete())

    def test_022_activity_loading(self):
        u1 = User(self.client, self.client.profile['id']).load()
        u2 = self.client.search_user(self.recipient_email)[0]
        post = u2.post('test')
        time.sleep(10)
        start = datetime.now(tz=timezone.utc).replace(tzinfo=timezone.utc) - timedelta(minutes=5)
        end = datetime.now(tz=timezone.utc).replace(tzinfo=timezone.utc)
        self.assertTrue(bool(u1.get_activity(start, end)), 'all activity list is empty')
        acts = u1.get_activity(start, end, [ActivityAction.POST])
        self.assertTrue(bool(acts), 'post activity list is empty')
        for act in acts:
            self.assertTrue(
                act.action == ActivityAction.POST, 'post activity list contains another action'
            )
        self.assertTrue(
            int(post.time_created.timestamp()) == int(acts[-1].time.timestamp()),
            'last post action time not match last post time',
        )

    def test_023_set_org_status(self):
        team = Team(self.client, self.org_id)
        team_member = team.member_info(self.client.profile['id'])
        new_status = f'Status test {datetime.now(tz=timezone.utc)}'
        self.assertTrue(team_member.add_status(new_status), 'add team status failed')
        self.assertTrue(team_member.get_last_status().status == new_status)

    def test_024_is_bot_meta(self):
        bot_post = self.bot.post_private_message_by_user_id(
            self.client.profile['id'], 'test message from bot'
        )
        chat = Chat(self.client, bot_post['chat_id'])
        post = chat.posts(bot_post['post_no'], bot_post['post_no'])[0]
        self.assertTrue(
            'is_bot' in post.meta.get('user', {}), 'meta.user.is_bot not found in received post'
        )
        self.assertTrue(
            post.meta.get('user', {}).get('is_bot') is True,
            'meta.user.is_bot is not True for post from bot',
        )

    def _poll_create(self, question: str, mode: str, anonymous: bool, options: List[str]) -> 'Poll':
        chat = Chat(self.client, self.recipient_chat_id)
        poll = Poll.create(chat, question=question, mode=mode, anonymous=anonymous, options=options)
        self.assertEqual(poll.question, question, 'poll question')
        self.assertEqual(poll.mode, mode, 'poll mode')
        self.assertEqual(poll.anonymous, anonymous, 'poll anonymous')
        self.assertEqual(poll.chat_id, self.recipient_chat_id, 'poll chat_id')
        self.assertListEqual([opt.text for opt in poll.options], options)
        return poll

    def test_025_single_poll(self):
        question = 'test anonymous single poll?'
        options = ['option 1', 'option 2', 'option 3']
        poll = self._poll_create(question, mode='one', anonymous=True, options=options)
        poll = poll.vote(0)
        self.assertEqual([opt.count for opt in poll.options], [1, 0, 0], 'poll vote')
        poll = poll.retract()
        self.assertTrue(all(opt.count == 0 for opt in poll.options), 'poll retract')
        exception = False
        try:
            poll.vote(4)
        except PararamioValidationException:
            exception = True
        self.assertTrue(exception, 'poll vote with incorrect option')
        exception = False
        try:
            poll.vote_multi([1, 2])
        except PararamioValidationException:
            exception = True
        self.assertTrue(exception, 'multi vote on single option poll')

    def test_026_multi_poll(self):
        question = 'test anonymous multi poll?'
        options = ['option 1', 'option 2', 'option 3']
        poll = self._poll_create(question, mode='more', anonymous=False, options=options)
        poll = poll.vote(1)
        self.assertEqual([opt.count for opt in poll.options], [0, 1, 0], 'poll vote')
        poll = poll.retract()
        self.assertTrue(all(opt.count == 0 for opt in poll.options), 'poll retract')
        poll = poll.vote_multi([1, 2])
        self.assertEqual([opt.count for opt in poll.options], [0, 1, 1], 'poll multi vote')
        exception = False
        try:
            poll.vote(4)
        except PararamioValidationException:
            exception = True
        self.assertTrue(exception, 'poll vote with incorrect option')
        exception = False
        try:
            poll.vote_multi([0, 4])
        except PararamioValidationException:
            exception = True
        self.assertTrue(exception, 'poll multi vote with incorrect option')

    def test_027_test_file(self):
        d = {'size': 23050, 'type': 'image/png', 'filename': 'test.png'}
        file_path = os.path.join(RESOURCES_PATH, 'test.png')
        chat = Chat(self.client, self.recipient_chat_id)
        file = chat.upload_file(file=file_path)
        file.delete()
        self.assertDictEqual(
            {'size': file.size, 'type': file.mime_type, 'filename': file.filename}, d
        )
        with self.assertRaises(PararamioValidationException):
            with open(file_path, 'rb') as f:
                chat.upload_file(file=f)
        with open(file_path, 'rb') as f:
            file2 = chat.upload_file(file=f, filename=d['filename'])
        file2.delete()
        self.assertDictEqual(
            {'size': file2.size, 'type': file2.mime_type, 'filename': file2.filename}, d
        )

    def test_028_test_mentions(self):
        u = self.client.search_user(self.recipient_email)[0]
        post = u.get_pm_thread().post('test @' + self.client2.profile.get('unique_name'))
        ## TODO: Uncomment this when the bug is fixed
        # self.assertTrue(post.is_mention)
        # self.assertTrue(post.mentions[0]['name'] == 'tamtambottester2')
        # self.assertTrue(post.mentions[0]['id'] == self.client.search_user('tamtambottester2')[0].id)
        # self.assertTrue(post.mentions[0]['value'] == '@tamtambottester2')

    def test_029_test_group_operations(self):
        grp = Group.create(self.client, organization_id=self.org_id, name='test_group')
        grp.delete()

    @classmethod
    def tearDownClass(cls):
        if os.path.exists('test.cookie'):
            os.remove('test.cookie')
        if os.path.exists('test.cookie2'):
            os.remove('test.cookie2')


def suite():
    return unittest.TestLoader().loadTestsFromTestCase(PararamioClientTest)


if __name__ == '__main__':
    unittest.main()



import sys
from http.cookiejar import CookieJar
from typing import Callable, Dict, Tuple, TypeVar, Union, TypedDict, Optional, Any

if sys.version_info >= (3, 11):
    # noinspection PyUnresolvedReferences
    from typing import NotRequired

    # noinspection PyUnresolvedReferences
    class ProfileTypeT(TypedDict):
        unique_name: str
        id: int
        info: Optional[str]
        find_strict: bool
        name: str
        is_google: bool
        two_step_enabled: bool
        has_password: bool
        phoneconfirmed: bool
        email: str
        phonenumber: NotRequired[Optional[str]]

    # noinspection PyUnresolvedReferences
    class TextParsedT(TypedDict):
        type: str
        value: str
        name: NotRequired[str]
        id: NotRequired[int]

    # noinspection PyUnresolvedReferences
    class PostMetaUserT(TypedDict):
        id: int
        name: str
        unique_name: str
        is_bot: NotRequired[bool]

else:

    class ProfileTypeT(TypedDict):
        unique_name: str
        id: int
        info: Optional[str]
        find_strict: bool
        name: str
        is_google: bool
        two_step_enabled: bool
        has_password: bool
        phoneconfirmed: bool
        email: str
        phonenumber: Optional[str]

    class TextParsedT(TypedDict):
        type: str
        value: str
        name: Optional[str]
        id: Optional[int]

    class PostMetaUserT(TypedDict):
        id: int
        name: str
        unique_name: str
        is_bot: Optional[bool]


class BotProfileT(TypedDict):
    id: int
    active: bool
    deleted: bool
    email: Optional[str]
    find_strict: bool
    has_password: bool
    info: Optional[str]
    info_parsed: list
    info_chat: int
    is_bot: bool
    is_google: bool
    name: str
    name_trans: str
    unique_name: str
    organizations: list
    phoneconfirmed: bool
    phonenumber: Optional[str]
    time_created: str
    time_updated: str
    two_step_enabled: bool


class PostMetaFileT(TypedDict):
    name: str
    guid: str
    size: int
    mime_type: str
    origin: Tuple[int, int]


class MetaReplyT(TypedDict):
    text: str
    user_id: int
    user_name: str
    in_thread_no: int


class PostMetaThreadT(TypedDict):
    title: str


class PostMetaT(TypedDict):
    user: PostMetaUserT
    thread: PostMetaThreadT
    file: PostMetaFileT
    reply: MetaReplyT


class PostMention(TypedDict):
    id: int
    name: str
    value: str


FormatterT = Dict[str, Callable[[Dict[str, Any], str], Any]]
CookieJarT = TypeVar('CookieJarT', bound=CookieJar)
QuoteRangeT = Dict[str, Union[str, int]]
HeaderLikeT = Dict[str, str]
SecondStepFnT = Callable[[CookieJar, Dict[str, str], str], Tuple[bool, Dict[str, str]]]



import logging
import mimetypes
import os
from http.cookiejar import CookieJar, FileCookieJar, LoadError, MozillaCookieJar
from io import BytesIO
from pathlib import Path
from typing import Any, BinaryIO, Callable, cast, Dict, Iterable, List, Optional, Tuple, Union

from pararamio._types import ProfileTypeT, SecondStepFnT
from pararamio.chat import Chat
from pararamio.constants import XSRF_HEADER_NAME
from pararamio.exceptions import (
    PararamioAuthenticationException,
    PararamioHTTPRequestException,
    PararamioValidationException,
)
from pararamio.file import File
from pararamio.group import Group
from pararamio.post import Post
from pararamio.user import User, UserSearchResult
from pararamio.utils.authentication import (
    authenticate,
    do_second_step,
    do_second_step_with_code,
    get_xsrf_token,
)
from pararamio.utils.helpers import (
    check_login_opts,
    get_empty_vars,
    lazy_loader,
    unescape_dict,
)
from pararamio.utils.requests import (
    api_request,
    delete_file,
    download_file,
    xupload_file,
)

__all__ = ('Pararamio',)
log = logging.getLogger('pararamio.client')


class Pararamio:
    """Pararamio client class.

    This class provides a client interface for interacting with the Pararamio API.

    Parameters:
        login: Optional string for the login name.
        password: Optional string for the password.
        key: Optional string for an authentication key.
        cookie: Optional CookieJar object for handling cookies.
        cookie_path: Optional string specifying the path to the cookie file.
        ignore_broken_cookie: Boolean flag to ignore broken cookie errors if set to True.
    """

    _login: Optional[str]
    _password: Optional[str]
    _key: Optional[str]
    _authenticated: bool
    _cookie: Union[CookieJar, FileCookieJar]
    __profile: Optional[ProfileTypeT]
    __headers: Dict[str, str]
    __user: dict

    def __init__(
        self,
        login: Optional[str] = None,
        password: Optional[str] = None,
        key: Optional[str] = None,
        cookie: Optional[CookieJar] = None,
        cookie_path: Optional[str] = None,
        ignore_broken_cookie: bool = False,
    ):
        self._login = login
        self._password = password
        self._key = key
        self.__headers = {}
        self.__profile = None
        self._authenticated = False
        if cookie is not None:
            self._cookie = cookie
        elif cookie_path is not None:
            self._cookie = MozillaCookieJar(cookie_path)
            if os.path.exists(cookie_path):
                if not os.path.isfile(cookie_path):
                    raise OSError(f'path {cookie_path} is directory')
                if not os.access(cookie_path, os.R_OK):
                    raise OSError(f'file {cookie_path} is not readable')
                if not os.access(cookie_path, os.W_OK):
                    raise OSError(f'file {cookie_path} is not writable')
                try:
                    self._cookie.load(ignore_discard=True)
                    self._authenticated = True
                except LoadError as e:
                    log.error('failed to load cookie from file %s', cookie_path)
                    if not ignore_broken_cookie:
                        raise OSError(e) from e
        else:
            self._cookie = CookieJar()
        for cj in self._cookie:
            if cj.name == '_xsrf':
                self.__headers[XSRF_HEADER_NAME] = str(cj.value)
                break

    @property
    def cookies(self) -> Union[CookieJar, FileCookieJar]:
        """

        Property for retrieving the cookie jar containing authentication cookies.

        Checks if the user is authenticated, and if not, performs the authentication process first.
        Once authenticated, returns the cookie jar.

        Returns:
            Union[CookieJar, FileCookieJar]: The cookie jar containing authentication cookies.
        """
        if not self._authenticated:
            self.authenticate()
        return self._cookie

    @property
    def headers(self) -> Dict[str, str]:
        """
        @property
        def headers(self) -> Dict[str, str]:
            Checks if the user is authenticated, performs authentication if not authenticated,
            and returns the headers.

            Returns:
                Dict[str, str]: The headers to be used in the request.
        """
        if not self._authenticated:
            self.authenticate()
        return self.__headers

    def _save_cookie(self) -> None:
        """
        _save_cookie:
            Saves the cookies in the FileCookieJar instance, if applicable.
            Ensures that cookies are saved persistently by ignoring the discard attribute.
        """
        if isinstance(self._cookie, FileCookieJar):
            self._cookie.save(ignore_discard=True)

    def _profile(self, raise_on_error: bool = False) -> 'ProfileTypeT':
        """

        Fetches the user profile data from the API.

        Parameters:
        - raise_on_error (bool): If set to True, an error will be raised in case of a failure.
                                 Defaults to False.

        Returns:
        - ProfileTypeT: The unescaped user profile data retrieved from the API.

        """
        return cast(
            'ProfileTypeT',
            unescape_dict(
                self.api_get('/user/me', raise_on_error=raise_on_error),
                keys=['name'],
            ),
        )

    def _do_auth(
        self,
        login: str,
        password: str,
        cookie_jar: CookieJar,
        headers: Dict[str, str],
        second_step_fn: SecondStepFnT,
        second_step_arg: str,
    ) -> None:
        """
        Authenticate the user and set necessary headers for future requests.

        Args:
            login (str): The user's login name.
            password (str): The user's password.
            cookie_jar (CookieJar): The cookie jar to store cookies.
            headers (Dict[str, str]): The headers to be included in the request.
            second_step_fn (SecondStepFnT): The function to handle
                                            the second step of authentication if required.
            second_step_arg (str): An argument for the second step function.

        Returns:
            None

        Sets:
            self._authenticated (bool): True if authentication is successful, False otherwise.
            self.__headers[XSRF_HEADER_NAME] (str): The XSRF token if authentication is successful.
        """
        self._authenticated, _, xsrf = authenticate(
            login, password, cookie_jar, headers, second_step_fn, second_step_arg
        )
        if self._authenticated:
            self.__headers[XSRF_HEADER_NAME] = xsrf
            self._save_cookie()

    def _authenticate(
        self,
        second_step_fn: SecondStepFnT,
        second_step_arg: str,
        login: Optional[str] = None,
        password: Optional[str] = None,
    ) -> bool:
        """
        Authenticate the user with the provided login and password,
        performing a secondary step if necessary.

        Arguments:
        second_step_fn: Function to execute for the second step of the authentication process
        second_step_arg: Argument to pass to the second step function
        login: Optional login name. If not provided,
               it will use login stored within the class instance.
        password: Optional password. If not provided,
                  it will use the password stored within the class instance.

        Returns:
        bool: True if authentication is successful, False otherwise

        Raises:
        PararamioAuthenticationException: If login or password is not provided or empty

        Exceptions:
        PararamioHTTPRequestException:
                        Raised if there is an error during the HTTP request in the profile check.

        """
        login = login or self._login or ''
        password = password or self._password or ''
        if not check_login_opts(login, password):
            raise PararamioAuthenticationException(
                f'{get_empty_vars(login=login, password=password)} must be set and not empty'
            )
        if not self._cookie:
            self._do_auth(
                login,
                password,
                self._cookie,
                self.__headers,
                second_step_fn,
                second_step_arg,
            )
        try:
            self._authenticated = True
            self._profile(raise_on_error=True)
        except PararamioHTTPRequestException:
            self._authenticated = False
            self._do_auth(
                login,
                password,
                self._cookie,
                self.__headers,
                second_step_fn,
                second_step_arg,
            )
        return self._authenticated

    def authenticate(
        self,
        login: Optional[str] = None,
        password: Optional[str] = None,
        key: Optional[str] = None,
    ) -> bool:
        """
        Authenticate a user using either a login and password or a key.

        This method attempts to authenticate a user through provided login credentials
        or a predefined key. If the key is not provided, it will use the instance key
        stored in `self._key`.

        Args:
            login (str, optional): The user's login name. Defaults to None.
            password (str, optional): The user's password. Defaults to None.
            key (str, optional): A predefined key for authentication. Defaults to None.

        Returns:
            bool: True if authentication is successful, False otherwise.

        Raises:
            PararamioAuthenticationException: If no key is provided.

        """
        key = key or self._key
        if not key:
            raise PararamioAuthenticationException('key must be set and not empty')
        return self._authenticate(do_second_step, key, login, password)

    def authenticate_with_code(
        self,
        code: str,
        login: Optional[str] = None,
        password: Optional[str] = None,
    ) -> bool:
        """

        Authenticates a user using a provided code and optionally login and password.

        Parameters:
          code (str): The authentication code. Must be set and not empty.
          login (str, optional): The user login. Default is None.
          password (str, optional): The user password. Default is None.

        Returns:
          bool: True if authentication is successful, otherwise raises an exception.

        Raises:
          PararamioAuthenticationException: If the code is not provided or is empty.
        """
        if not code:
            raise PararamioAuthenticationException('code must be set and not empty')
        return self._authenticate(do_second_step_with_code, code, login, password)

    def _api_request(
        self,
        url: str,
        method: str = 'GET',
        data: Optional[dict] = None,
        callback: Callable = lambda rsp: rsp,
        raise_on_error: bool = False,
    ) -> Any:
        """
        Performs an authenticated API request with XSRF token management and error handling.

        Args:
            url (str): The API endpoint URL to which the request is made.
            method (str): The HTTP method to use for the request. Defaults to 'GET'.
            data (Optional[dict]): The data payload for the request, if applicable.
                                   Defaults to None.
            callback (Callable): A callback function to process the response.
                                 Defaults to a lambda that returns the response.
            raise_on_error (bool): Flag to determine if exceptions should be raised.
                                   Defaults to False.

        Returns:
            Any: The result of the callback processing on the API request response.

        Raises:
            PararamioHTTPRequestException:
                                         If an HTTP error occurs and raise_on_error is set to True.

        Notes:
            - The function ensures that the user is authenticated before making the request.
            - Manages the XSRF token by retrieving and saving it as needed.
            - Handles specific error cases by attempting re-authentication or
              renewing the XSRF token.
        """
        if not self._authenticated:
            self.authenticate()
        if not self.__headers.get(XSRF_HEADER_NAME, None):
            self.__headers[XSRF_HEADER_NAME] = get_xsrf_token(self._cookie)
            self._save_cookie()
        try:
            return callback(
                api_request(url, method, data, cookie_jar=self._cookie, headers=self.__headers)
            )
        except PararamioHTTPRequestException as e:
            if raise_on_error:
                raise
            if e.code == 401:
                self._authenticated = False
                return self._api_request(
                    url=url,
                    method=method,
                    data=data,
                    callback=callback,
                    raise_on_error=True,
                )
            message = e.message
            if message == 'xsrf':
                log.info('xsrf is expire, invalid or was not set, trying to get new one')
                self.__headers[XSRF_HEADER_NAME] = ''
                return self._api_request(
                    url=url,
                    method=method,
                    data=data,
                    callback=callback,
                    raise_on_error=True,
                )
            raise

    def api_get(self, url: str, raise_on_error: bool = False) -> dict:
        """

        Handles HTTP GET requests to the specified API endpoint.

        Arguments:
        url (str): The URL of the API endpoint.

        raise_on_error (bool): If set to True, an exception will be raised
                               if the API response indicates an error. Defaults to False.

        Returns:
        dict: The JSON response from the API, parsed into a Python dictionary.

        """
        return self._api_request(url, raise_on_error=raise_on_error)

    def api_post(
        self,
        url: str,
        data: Optional[Dict[Any, Any]] = None,
        raise_on_error: bool = False,
    ) -> dict:
        """
        Sends a POST request to the specified URL with the given data.

        Parameters:
        url (str): The endpoint URL where the POST request should be sent.
        data (Optional[Dict[Any, Any]], optional): The payload to be sent in the POST request body.
                                                   Defaults to None.
        raise_on_error (bool, optional): Whether to raise an exception if the request fails.
                                         Defaults to False.

        Returns:
        dict: The response from the server as a dictionary.
        """
        return self._api_request(url, method='POST', data=data, raise_on_error=raise_on_error)

    def api_put(
        self,
        url: str,
        data: Optional[Dict[Any, Any]] = None,
        raise_on_error: bool = False,
    ) -> dict:
        """
        Sends a PUT request to the specified URL with the provided data.

        Parameters:
        - url: The URL to send the PUT request to.
        - data: Optional dictionary containing the data to include in the request body.
        - raise_on_error: Boolean flag indicating whether to raise an exception
                          if the request results in an error.

        Returns:
        A dictionary containing the server's response to the PUT request.
        """
        return self._api_request(url, method='PUT', data=data, raise_on_error=raise_on_error)

    def api_delete(
        self,
        url: str,
        data: Optional[Dict[Any, Any]] = None,
        raise_on_error: bool = False,
    ) -> dict:
        """
        Sends a DELETE request to the specified URL with optional data.

        Parameters:
        url (str): The URL to send the DELETE request to.
        data (Optional[Dict[Any, Any]], optional): Optional payload to include in the request.
        raise_on_error (bool, optional): Determines whether an exception should be raised
                                         on request failure.

        Returns:
        dict: The response from the API request.
        """
        return self._api_request(url, method='DELETE', data=data, raise_on_error=raise_on_error)

    def _upload_file(
        self,
        file: BinaryIO,
        chat_id: int,
        filename: Optional[str] = None,
        type_: Optional[str] = None,
        organization_id: Optional[int] = None,
        reply_no: Optional[int] = None,
        quote_range: Optional[str] = None,
    ) -> Tuple[dict, dict]:
        """
        _upload_file is a method for uploading a file to a specified chat or organization.

        Arguments:
            file: A binary stream of the file to be uploaded.
            chat_id: The ID of the chat where the file will be uploaded.
            filename: An optional parameter that specifies the name of the file.
            type_: An optional parameter that specifies the type of file being uploaded.
                   If not provided, it will be inferred from the filename.
            organization_id: An optional parameter that specifies the ID of the organization
                             if the file is an organization avatar.
            reply_no: An optional parameter that specifies the reply number
                      associated with the file.
            quote_range: An optional parameter that specifies the range
                         of quotes associated with the file.

        Returns:
            A tuple containing a dictionary with the response from the xupload_file function
            and a dictionary of the fields used during the upload.

        Raises:
            PararamioValidationException: If filename is not set when type is None,
            or if organization_id is not set when type is organization_avatar,
            or if chat_id is not set when type is chat_avatar.

        Notes:
            This method ensures that the necessary headers and
            tokens are set before attempting the file upload.
        """
        if type_ is None and not filename:
            raise PararamioValidationException('filename must be set when type is None')
        if not self._authenticated:
            self.authenticate()
        if not self.__headers.get(XSRF_HEADER_NAME, None):
            self.__headers[XSRF_HEADER_NAME] = get_xsrf_token(self._cookie)
        if type_ == 'organization_avatar' and organization_id is None:
            raise PararamioValidationException(
                'organization_id must be set when type is organization_avatar'
            )
        if type_ == 'chat_avatar' and chat_id is None:
            raise PararamioValidationException('chat_id must be set when type is chat_avatar')
        if type_ is None:
            if not mimetypes.inited:
                mimetypes.init(files=os.environ.get('PARARAMIO_MIME_TYPES_PATH', None))
            type_ = mimetypes.guess_type(filename)[0]  # type: ignore[arg-type]
        file.seek(0, os.SEEK_END)
        file_size = file.tell()
        file.seek(0, 0)
        fields: List[Tuple[str, Union[str, int, None]]] = [
            ('type', type_),
            ('filename', filename),
            ('size', file_size),
            ('chat_id', chat_id),
            ('organization_id', organization_id),
            ('reply_no', reply_no),
            ('quote_range', quote_range),
        ]
        return xupload_file(
            fp=file, fields=fields, headers=self.__headers, cookie_jar=self._cookie
        ), dict(fields)

    def upload_file(
        self,
        file: Union[str, BytesIO, Path],
        chat_id: int,
        filename: Optional[str] = None,
        reply_no: Optional[int] = None,
        quote_range: Optional[str] = None,
    ) -> File:
        """

        Uploads a file to a specified chat.

        Args:
            file (Union[str, BytesIO, Path]): The file to be uploaded.
                                              Can be a file path as a string or Path object,
                                              or a file-like object in memory.
            chat_id (int): The identifier of the chat where the file will be uploaded.
            filename (Optional[str], optional): The name of the file.
                                                If not provided, the basename of the file path is
                                                used. Defaults to None.
            reply_no (Optional[int], optional): The message number in the chat to which this
                                                file is a reply. Defaults to None.
            quote_range (Optional[str], optional): The range of the message to quote.
                                                   Defaults to None.

        Returns:
            File: An object representing the uploaded file,
                  including its unique identifier, MIME type, and other metadata.
        """
        if isinstance(file, (str, Path)):
            filename = filename or os.path.basename(file)
            with open(file, 'rb') as f:
                res, extra = self._upload_file(
                    file=f,
                    chat_id=chat_id,
                    filename=filename,
                    reply_no=reply_no,
                    quote_range=quote_range,
                )
        else:
            res, extra = self._upload_file(
                file=file,
                chat_id=chat_id,
                filename=filename,
                reply_no=reply_no,
                quote_range=quote_range,
            )
        return File(self, guid=res['guid'], mime_type=extra['type'], **extra)

    def delete_file(self, guid: str) -> dict:
        """
        Deletes a file identified by the provided GUID.

        Args:
            guid (str): The globally unique identifier of the file to be deleted.

        Returns:
            dict: The result of the deletion operation.

        """
        return delete_file(guid, headers=self.__headers, cookie_jar=self._cookie)

    def download_file(self, guid: str, filename: str) -> BytesIO:
        """
        Downloads and returns a file as a BytesIO object given its unique identifier and filename.

        Args:
            guid (str): The unique identifier of the file to be downloaded.
            filename (str): The name of the file to be downloaded.

        Returns:
            BytesIO: A BytesIO object containing the downloaded file content.
        """
        return download_file(guid, filename, headers=self.__headers, cookie_jar=self._cookie)

    @property
    def profile(self) -> 'ProfileTypeT':
        """

        Provides access to the profile property. If the profile is not
        yet initialized, this method will initialize it by calling the
        _profile method.

        Returns:
            ProfileTypeT: The profile object.
        """
        if not self.__profile:
            self.__profile = self._profile()
        return self.__profile

    def search_user(self, query: str) -> List[UserSearchResult]:
        """
        search_user(query: str) -> List[User]

        Searches for users based on the given query string.

        Parameters:
        query (str): The search query used to find matching users.

        Returns:
        List[User]: A list of User objects that match the search query.
        """
        return User.search(self, query)

    def search_group(self, query: str) -> List[Group]:
        """
        Performs a search for groups based on a given query string.

        Arguments:
        query (str): The search term used to find matching groups.

        Returns:
        List[Group]: A list of Group objects that match the search criteria.
        """
        return Group.search(self, query)

    def search_posts(
        self,
        query: str,
        order_type: str = 'time',
        page: int = 1,
        chat_id: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> Tuple[int, Iterable[Post]]:
        """

        search_posts searches for posts based on a given query and various optional parameters.

        Arguments:
        - query: The search term used to find posts.
        - order_type: Specifies the order of the search results. Default is 'time'.
        - page: The page number of the search results to retrieve. Default is 1.
        - chat_id: Optional ID of the chat to search within. If None, search in all chats.
        - limit: The maximum number of posts to return. If None, use the default limit.

        Returns:
        - A tuple containing the total number of posts matching
          the search query and an iterable of Post objects.
        """
        return Chat.post_search(
            self, query, order_type=order_type, page=page, chat_id=chat_id, limit=limit
        )

    def list_chats(self) -> Iterable:
        """
        Returns an iterable that yields chat objects in a lazy-loading manner.
        The chats are fetched from the server using the specified URL and are returned in batches.

        Returns:
            Iterable: An iterable that yields chat objects.
        """
        url = '/core/chat/sync'
        chats_per_load = 50
        ids = self.api_get(url).get('chats', [])
        return lazy_loader(self, ids, Chat.load_chats, per_load=chats_per_load)

    def get_groups_by_ids(self, ids: List[int]) -> Iterable[Group]:
        """

        Fetches a list of Group objects based on provided group ids.

        Args:
         ids (List[int]): A list of group IDs to fetch.

        Returns:
         Iterable[Group]: An iterable of Group objects corresponding to the provided IDs.

        """
        url = '/core/group?ids=' + ','.join(map(str, ids))
        return [Group(self, **data) for data in self.api_get(url).get('groups', [])]

    def get_users_by_ids(self, ids: List[int]) -> List[User]:
        """
        Retrieve a list of User objects by their IDs.

        Arguments:
        ids -- a list of user ID integers.

        Returns:
        A list of User objects corresponding to the provided IDs.
        """
        url = '/user/list?ids=' + ','.join(map(str, ids))
        return [
            User(self, **unescape_dict(data, ['name']))
            for data in self.api_get(url).get('users', [])
        ]

    def post_private_message_by_user_email(self, email: str, text: str) -> Post:
        """

        Posts a private message to a user identified by their email address.

        :param email: The email address of the user to whom the message will be sent.
        :type email: str
        :param text: The content of the message to be posted.
        :type text: str
        :return: A Post object representing the posted message.
        :rtype: Post
        """
        url = '/msg/post/private'
        resp = self._api_request(url, method='POST', data={'text': text, 'user_email': email})
        return Post(Chat(self, resp['chat_id']), resp['post_no'])

    def post_private_message_by_user_id(self, user_id: int, text: str) -> Post:
        """
        Send a private message to a specific user.

        Parameters:
        user_id (int): The ID of the user to whom the message will be sent.
        text (str): The content of the message to be sent.

        Returns:
        Post: The Post object containing information about the sent message.
        """
        url = '/msg/post/private'
        resp = self._api_request(url, method='POST', data={'text': text, 'user_id': user_id})
        return Post(Chat(self, resp['chat_id']), resp['post_no'])

    def post_private_message_by_user_unique_name(self, unique_name: str, text: str) -> Post:
        """
        Post a private message to a user identified by their unique name.

        Parameters:
        unique_name (str): The unique name of the user to whom the private message is to be sent.
        text (str): The content of the private message.

        Returns:
        Post: An instance of the Post class representing the posted message.
        """
        url = '/msg/post/private'
        resp = self._api_request(
            url, method='POST', data={'text': text, 'user_unique_name': unique_name}
        )
        return Post(Chat(self, resp['chat_id']), resp['post_no'])



import codecs
import json
import logging
import mimetypes
import os
from http.client import HTTPResponse
from io import BytesIO
from json import JSONDecodeError
from typing import BinaryIO, List, Optional, Tuple, Union
from urllib.error import HTTPError
from urllib.request import HTTPCookieProcessor, Request, build_opener

from pararamio._types import CookieJarT, HeaderLikeT
from pararamio.constants import (
    BASE_API_URL,
    FILE_UPLOAD_URL,
    REQUEST_TIMEOUT as TIMEOUT,
    UPLOAD_TIMEOUT,
    VERSION,
)
from pararamio.exceptions import PararamioHTTPRequestException


__all__ = (
    'api_request',
    'bot_request',
    'upload_file',
    'xupload_file',
    'delete_file',
    'download_file',
    'raw_api_request',
)
log = logging.getLogger('pararamio')
UA_HEADER = f'pararamio lib version {VERSION}'
DEFAULT_HEADERS = {
    'Content-type': 'application/json',
    'Accept': 'application/json',
    'User-agent': UA_HEADER,
}
writer = codecs.lookup('utf-8')[3]


def multipart_encode(
    fd: BinaryIO,
    fields: Optional[List[Tuple[str, Union[str, None, int]]]] = None,
    boundary: Optional[str] = None,
    form_field_name: str = 'data',
    filename: Optional[str] = None,
    content_type: Optional[str] = None,
) -> bytes:
    """
    Encodes a file and additional fields into a multipart/form-data payload.

    Args:
        fd: A file-like object opened in binary mode that is to be included in the payload.
        fields: An optional list of tuples representing additional form fields,
                with each tuple containing a field name and its value.
        boundary: An optional string used to separate parts of the multipart message.
                  If not provided, a default boundary ('FORM-BOUNDARY') is used.
        form_field_name: The name of the form field for the file being uploaded. Defaults to 'data'.
        filename: An optional string representing the filename for the file being uploaded.
                  If not provided, the name is derived from the file-like object.
        content_type: An optional string representing the content type of the file being uploaded.
                      If not provided, the content type will be guessed from the filename.

    Returns:
        A bytes' object representing the encoded multipart/form-data payload.
    """
    if fields is None:
        fields = []
    if boundary is None:
        boundary = 'FORM-BOUNDARY'
    body = BytesIO()

    def write(text: str):
        nonlocal body
        writer(body).write(text)

    if fields:
        for key, value in fields:
            if value is None:
                continue
            write(f'--{boundary}\r\n')
            write(f'Content-Disposition: form-data; name="{key}"')
            write(f'\r\n\r\n{value}\r\n')
    if not filename:
        filename = os.path.basename(fd.name)
    if not content_type:
        content_type = mimetypes.guess_type(filename)[0] or 'application/octet-stream'
    fd.seek(0)
    write(f'--{boundary}\r\n')
    write(f'Content-Disposition: form-data; name="{form_field_name}"; filename="{filename}"\r\n')
    write(f'Content-Type: {content_type}\r\n\r\n')
    body.write(fd.read())
    write(f'\r\n--{boundary}--\r\n\r\n')
    return body.getvalue()


def bot_request(
    url: str,
    key: str,
    method: str = 'GET',
    data: Optional[dict] = None,
    headers: Optional[dict] = None,
    timeout: int = TIMEOUT,
):
    """

    Sends a request to a bot API endpoint with the specified parameters.

    Parameters:
    url (str): The endpoint URL of the bot API.
    key (str): The API token for authentication.
    method (str): The HTTP method to use for the request. Defaults to 'GET'.
    data (Optional[dict]): The data payload for the request. Defaults to None.
    headers (Optional[dict]): Additional headers to include in the request. Defaults to None.
    timeout (int): The timeout setting for the request. Defaults to TIMEOUT.

    Returns:
    Response object from the API request.
    """
    _headers = {'X-APIToken': key, **DEFAULT_HEADERS}
    if headers:
        _headers = {**_headers, **headers}
    return api_request(url=url, method=method, data=data, headers=_headers, timeout=timeout)


def _base_request(
    url: str,
    method: str = 'GET',
    data: Optional[bytes] = None,
    headers: Optional[dict] = None,
    cookie_jar: Optional[CookieJarT] = None,
    timeout: int = TIMEOUT,
) -> HTTPResponse:
    """
    Sends an HTTP request and returns an HTTPResponse object.

    Parameters:
        url (str): The URL endpoint to which the request is sent.
        method (str): The HTTP method to use for the request (default is 'GET').
        data (Optional[bytes]): The payload to include in the request body (default is None).
        headers (Optional[dict]): A dictionary of headers to include in the request
                                  (default is None).
        cookie_jar (Optional[CookieJarT]): A cookie jar to manage cookies for the request
                                           (default is None).
        timeout (int): The timeout for the request in seconds (TIMEOUT defines default).

    Returns:
        HTTPResponse: The response object containing the server's response to the HTTP request.

    Raises:
        PararamioHTTPRequestException: An exception is raised if there is
                                       an issue with the HTTP request.
    """
    _url = f'{BASE_API_URL}{url}'
    _headers = DEFAULT_HEADERS
    if headers:
        _headers = {**_headers, **headers}
    opener = build_opener(*[HTTPCookieProcessor(cookie_jar)] if cookie_jar is not None else [])
    _data = None
    if data:
        _data = data
    rq = Request(_url, _data, method=method, headers=_headers)
    log.debug('%s - %s - %s - %s - %s', _url, method, data, headers, cookie_jar)
    try:
        return opener.open(rq, timeout=timeout)
    except HTTPError as e:
        log.error('%s - %s - %s', _url, method, e)
        # noinspection PyUnresolvedReferences
        raise PararamioHTTPRequestException(e.filename, e.code, e.msg, e.hdrs, e.fp) from e


def _base_file_request(
    url: str,
    method='GET',
    data: Optional[bytes] = None,
    headers: Optional[HeaderLikeT] = None,
    cookie_jar: Optional[CookieJarT] = None,  # type: ignore
    timeout: int = TIMEOUT,
) -> BytesIO:
    """
    Performs a file request to the specified URL with the given parameters.

    Arguments:
        url (str): The URL endpoint to send the request to.
        method (str, optional): The HTTP method to use for the request (default is 'GET').
        data (Optional[bytes], optional): The data to send in the request body (default is None).
        headers (Optional[HeaderLikeT], optional): The headers to include in the request
                                                   (default is None).
        cookie_jar (Optional[CookieJarT], optional): The cookie jar to use for managing cookies
                                                     (default is None).
        timeout (int, optional): The timeout duration for the request (default value is TIMEOUT).

    Returns:
        BytesIO: The response object from the file request.

    Raises:
        PararamioHTTPRequestException: If the request fails with an HTTP error code.
    """
    _url = f'{FILE_UPLOAD_URL}{url}'
    opener = build_opener(HTTPCookieProcessor(cookie_jar))
    if not headers:
        headers = {}
    rq = Request(_url, data, method=method, headers=headers)
    log.debug('%s - %s - %s - %s - %s', url, method, data, headers, cookie_jar)
    try:
        resp = opener.open(rq, timeout=timeout)
        if 200 >= resp.getcode() < 300:
            return resp
        raise PararamioHTTPRequestException(
            _url, resp.getcode(), 'Unknown error', resp.getheaders(), resp.fp
        )
    except HTTPError as e:
        log.error('%s - %s - %s', _url, method, e)
        # noinspection PyUnresolvedReferences
        raise PararamioHTTPRequestException(e.filename, e.code, e.msg, e.hdrs, e.fp) from e


def upload_file(
    fp: BinaryIO,
    perm: str,
    filename: Optional[str] = None,
    file_type=None,
    headers: Optional[HeaderLikeT] = None,
    cookie_jar: Optional[CookieJarT] = None,
    timeout: int = UPLOAD_TIMEOUT,
):
    """
    Upload a file to a pararam server with specified permissions and optional parameters.

    Args:
        fp (BinaryIO): A file-like object to be uploaded.
        perm (str): The permission level for the uploaded file.
        filename (Optional[str], optional): Optional filename used during upload. Defaults to None.
        file_type (optional): Optional MIME type of the file. Defaults to None.
        headers (Optional[HeaderLikeT], optional): Optional headers to include in the request.
        Defaults to None.
        cookie_jar (Optional[CookieJarT], optional): Optional cookie jar for maintaining session.
        Defaults to None.
        timeout (int, optional): Timeout duration for the upload request.
        Defaults to UPLOAD_TIMEOUT.

    Returns:
        dict: A dictionary containing the server's response to the file upload.

    The function constructs a multipart form data request with the file contents,
    sends the POST request to the server,
    and returns the parsed JSON response from the server.
    """
    url = f'/upload/{perm}'
    boundary = 'FORM-BOUNDARY'
    _headers = {
        'User-agent': UA_HEADER,
        **(headers or {}),
        'Accept': 'application/json',
        'Content-Type': f'multipart/form-data; boundary={boundary}',
    }
    data = multipart_encode(
        fp,
        boundary=boundary,
        form_field_name='file',
        filename=filename,
        content_type=file_type,
    )
    resp = _base_file_request(
        url,
        method='POST',
        data=data,
        headers=_headers,
        cookie_jar=cookie_jar,
        timeout=timeout,
    )
    return json.loads(resp.read())


def xupload_file(
    fp: BinaryIO,
    fields: List[Tuple[str, Union[str, None, int]]],
    headers: Optional[HeaderLikeT] = None,
    cookie_jar: Optional[CookieJarT] = None,
    timeout: int = UPLOAD_TIMEOUT,
) -> dict:
    """

    Uploads a file to the server using multipart/form-data encoding.

    Parameters:
        fp (BinaryIO): A binary file-like object to be uploaded.
        fields (List[Tuple[str, Union[str, None, int]]]): A list of tuples representing form fields
        to be included in the upload. Each tuple contains the field name and its value.
        headers (Optional[HeaderLikeT]): Additional headers to include in the request.
        cookie_jar (Optional[CookieJarT]): CookieJar object to handle cookies during the request.
        timeout (int): Timeout value for the request.

    Returns:
        dict: The server's response parsed as a JSON dictionary.
    """
    url = '/upload'
    boundary = 'FORM-BOUNDARY'
    _headers = {
        'User-agent': UA_HEADER,
        **(headers or {}),
        'Accept': 'application/json',
        'Content-Type': f'multipart/form-data; boundary={boundary}',
    }
    data = multipart_encode(fp, fields, boundary=boundary)
    resp = _base_file_request(
        url,
        method='POST',
        data=data,
        headers=_headers,
        cookie_jar=cookie_jar,
        timeout=timeout,
    )
    return json.loads(resp.read())


def delete_file(
    guid: str,
    headers: Optional[HeaderLikeT] = None,
    cookie_jar: Optional[CookieJarT] = None,
    timeout: int = TIMEOUT,
) -> dict:
    url = f'/delete/{guid}'
    resp = _base_file_request(
        url, method='DELETE', headers=headers, cookie_jar=cookie_jar, timeout=timeout
    )
    return json.loads(resp.read())


def download_file(
    guid: str,
    filename: str,
    headers: Optional[HeaderLikeT] = None,
    cookie_jar: Optional[CookieJarT] = None,
    timeout: int = TIMEOUT,
) -> BytesIO:
    url = f'/download/{guid}/{filename}'
    res = file_request(url, method='GET', headers=headers, cookie_jar=cookie_jar, timeout=timeout)
    return res


def file_request(
    url: str,
    method='GET',
    data: Optional[bytes] = None,
    headers: Optional[HeaderLikeT] = None,
    cookie_jar: Optional[CookieJarT] = None,
    timeout: int = TIMEOUT,
) -> BytesIO:
    _headers = DEFAULT_HEADERS
    if headers:
        _headers = {**_headers, **headers}
    return _base_file_request(
        url,
        method=method,
        data=data,
        headers=_headers,
        cookie_jar=cookie_jar,
        timeout=timeout,
    )


def raw_api_request(
    url: str,
    method: str = 'GET',
    data: Optional[bytes] = None,
    headers: Optional[HeaderLikeT] = None,
    cookie_jar: Optional[CookieJarT] = None,
    timeout: int = TIMEOUT,
) -> Tuple[dict, List]:
    resp = _base_request(url, method, data, headers, cookie_jar, timeout)
    if 200 >= resp.getcode() < 300:
        contents = resp.read()
        return json.loads(contents), resp.getheaders()
    return {}, []


def api_request(
    url: str,
    method: str = 'GET',
    data: Optional[dict] = None,
    headers: Optional[HeaderLikeT] = None,
    cookie_jar: Optional[CookieJarT] = None,
    timeout: int = TIMEOUT,
) -> dict:
    _data = None
    if data is not None:
        _data = str.encode(json.dumps(data), 'utf-8')
    resp = _base_request(url, method, _data, headers, cookie_jar, timeout)
    resp_code = resp.getcode()
    if resp_code == 204:
        return {}
    if 200 <= resp_code < 500:
        content = resp.read()
        try:
            return json.loads(content)
        except JSONDecodeError as e:
            log.error('%s - %s - %s', url, method, e)
            raise PararamioHTTPRequestException(
                url,
                resp.getcode(),
                'JSONDecodeError',
                resp.getheaders(),
                BytesIO(content),
            ) from e
    return {}



from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, TYPE_CHECKING, TypedDict
from pararamio.activity import Activity, ActivityAction

from pararamio.exceptions import PararamNotFound
from pararamio.utils.helpers import unescape_dict

from .chat import Chat
from ._base import BaseLoadedAttrPararamObject

if TYPE_CHECKING:
    from ._types import QuoteRangeT
    from pararamio.client import Pararamio
    from pararamio.post import Post
    from datetime import datetime

__all__ = ('User', 'UserSearchResult', 'UserInfoParsedItem')


@dataclass
class UserSearchResult:
    id: int
    avatar: Optional[str]
    name: str
    unique_name: str
    custom_name: Optional[str]
    time_created: str
    time_updated: str
    other_blocked: bool
    pm_thread_id: Optional[int]
    is_bot: bool
    user: 'User'

    @property
    def has_pm(self) -> bool:
        return self.pm_thread_id is not None

    def get_pm_thread(self) -> 'Chat':
        if self.pm_thread_id is not None:
            chat = Chat(self.user._client, self.pm_thread_id)
            return chat
        return Chat.create_private_chat(self.user._client, self.id)

    def post(
        self,
        text: str,
        quote_range: Optional['QuoteRangeT'] = None,
        reply_no: Optional[int] = None,
    ) -> 'Post':
        chat = self.get_pm_thread()
        return chat.post(text=text, quote_range=quote_range, reply_no=reply_no)


class UserInfoParsedItem(TypedDict):
    type: str
    value: str


INTERSECTION_KEYS = (
    'id',
    'name',
    'unique_name',
    'time_created',
    'time_updated',
    'is_bot',
)


class User(BaseLoadedAttrPararamObject):
    __slots__ = (
        '_data',
        '_client',
        'id',
        '_load_on_key_error',
    )
    id: int
    name: str
    name_trans: str
    info: str
    unique_name: str
    deleted: bool
    active: bool
    time_updated: str
    time_created: str
    is_bot: bool
    alias: Optional[None]
    timezone_offset_minutes: int
    owner_id: Optional[None]
    organizations: List[int]
    info_parsed: List[UserInfoParsedItem]
    _data: Dict[str, Any]
    _client: 'Pararamio'

    def __init__(self, client, id: int, load_on_key_error: bool = True, **kwargs):
        self._client = client
        self.id = id
        self._data = {'id': id, **kwargs}
        self._load_on_key_error = load_on_key_error

    def __eq__(self, other):
        if not isinstance(other, User):
            return id(other) == id(self)
        return self.id == other.id

    def load(self) -> 'User':
        resp = self._client.get_users_by_ids([self.id])
        if len(resp) != 1:
            raise PararamNotFound()
        self._data = next(iter(resp))._data
        return self

    @classmethod
    def load_users(cls, client, ids: List[int]) -> List['User']:
        return client.get_users_by_ids(ids)

    def post(
        self,
        text: str,
        quote_range: Optional['QuoteRangeT'] = None,
        reply_no: Optional[int] = None,
    ) -> 'Post':
        for res in self.search(self._client, self.unique_name):
            if res.unique_name == self.unique_name:
                return res.post(text=text, quote_range=quote_range, reply_no=reply_no)
        raise PararamNotFound(f'User {self.unique_name} not found')

    def __str__(self):
        if 'name' not in self._data:
            self.load()
        return self._data.get('name')

    @classmethod
    def search(cls, client: 'Pararamio', search_string: str) -> List[UserSearchResult]:
        url = f'/users?flt={search_string}'
        result: List[UserSearchResult] = []
        for response in client.api_get(url).get('users', []):
            data = unescape_dict(response, keys=['name'])
            data['user'] = cls(client, **{k: data[k] for k in INTERSECTION_KEYS})
            result.append(UserSearchResult(**data))
        return result

    def _activity_page_loader(self) -> Callable[..., Dict[str, Any]]:
        def loader(action: Optional[ActivityAction] = None, page: int = 1) -> Dict[str, Any]:
            action_ = action.value if action else ''
            url = f'/activity?user_id={self.id}&action={action_}&page={page}'
            return self._client.api_get(url)

        return loader

    def get_activity(
        self,
        start: 'datetime',
        end: 'datetime',
        actions: Optional[List[ActivityAction]] = None,
    ) -> List[Activity]:
        """get user activity

        :param start: start time
        :param end: end time
        :param actions: list of action types (all actions if None)
        :returns: activity list
        """
        return Activity.get_activity(self._activity_page_loader(), start, end, actions)
